# -*- coding:utf-8 -*-
"""두음법칙에 관련된 내용은
http://ko.wikipedia.org/wiki/%EB%91%90%EC%9D%8C_%EB%B2%95%EC%B9%99 를 참고.
"""
from .impl import translate

__author__ = "Sumin Byeon"
__email__ = "suminb@gmail.com"
__version__ = "0.13.3"
